/* page.c --- pagination routines live here
**
**  These routines are for simple report pagination.  The interface
**  consists of:
**
**      pg_open (FILE *fp, int pglen, int pgwidth, int headlen,
**               int taillen);
**      pg_head (int hdlineno, char *line, char sep);
**      pg_printf (format, arg1, ...,, argn);
**      pg_close ();
**
**  pg_open sets the FILE pointer to use, the page length and width,
**  the header length, and trailer length.
**  pg_head sets the hdlineno line to the contents of line.  Line can
**  contain chars of value sep, and will be right/center/left
**  justified.  They can also contain format strings as described in
**  cftime(3).  pg_printf calls vsprintf, and works like printf.
**  pg_close fills out the last page, and flushes the stream.  It does
**  not close it.
*/

/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <stdarg.h>

typedef struct {
    FILE    *fp;        /* FILE pointer we're using */
    int     lineno;     /* current lineno */
    char    **head;     /* header lines */
    int     sep;        /* seperator */
    int     pglen;      /* page length */
    int     pgwidth;    /* page width */
    int     nhead;      /* # of header lines */
    int     ntail;      /* # of tail lines */
} page_t;

static page_t P;
static char Buf[1024];

static void pg_buf (void);
static void pg_line (char *, char *);
static void pg_do_tailer (void);
static void pg_do_header (void);
static void copy (char *, char *);

/* pg_open --- initilize the pager */

int
pg_open (FILE *fp, int pglen, int pgwidth, int nhead, int ntail)
{
    int i;
    
    if (nhead+ntail+1 > pglen) {
	fprintf (stderr, "pg_open: page length too short for nhead and ntail\n");
	return (-1);
    }
	
    if (P.fp)
	pg_close ();
    
    P.fp = fp;
    P.lineno = 0;
    P.head = calloc (sizeof (char *), nhead);
    if (P.head == NULL) {
	fprintf (stderr, "pg_open: Out of memory!\n");
	return (-1);
    }
    for (i = 0; i < nhead; i++)
	P.head[i] = NULL;
	
    P.pglen = pglen;
    P.pgwidth = pgwidth;
    P.nhead = nhead;
    P.ntail = ntail;
    return (0);
}

/* pg_head --- specify a header line */

int
pg_head (int h, char *str, char sep)
{
    char *p;
    int nsep;
    
    if (h < 0 || h >= P.nhead) {
	fprintf (stderr, "pg_head: h out of range 0-%d\n", P.nhead-1);
	return (-1);
    }
    
    nsep = 0;
    for (p = str; *p; p++)
	if (*p == sep)
	    ++nsep;
    
    if (nsep > 2) {
	fprintf (stderr, "pg_head: too many seps in str\n");
	return (-1);
    }
    
    P.head[h] = strdup (str);
    if (P.head[h] == NULL) {
	fprintf (stderr, "pg_head: Out of mem!\n");
	return (-1);
    }
    
    P.sep = sep;
    
    return (0);
}

/* pg_printf --- like printf, but handles pages */

int
pg_printf (char *format, ...)
{
    va_list ap;

    va_start (ap, format);
    vsprintf (Buf, format, ap);
    va_end (ap);
    
    if ((int) strlen (Buf) > P.pgwidth) {
	Buf[P.pgwidth] = '\n';
	Buf[P.pgwidth+1] = 0;
    }
    
    pg_buf ();
    
    return (0);
}

/* pg_close --- finish the page, and dismantle everything */

int
pg_close ()
{
    int i;
    
    if (P.lineno != 0)
	while (P.lineno < P.pglen) {
	    putc ('\n', P.fp);
	    ++P.lineno;
	}
	
    fflush (P.fp);
    
    for (i = 0; i < P.nhead; i++)
	if (P.head[i])
	    free (P.head[i]);
    
    if (P.head)
	free (P.head);
    P.head = NULL;
    P.fp = NULL;
    P.lineno = 0;
    P.pglen = 0;
    P.pgwidth = 0;
    P.nhead = 0;
    P.ntail = 0;
}

/* pg_buf --- Buf may contain many lines */

static void
pg_buf ()
{
    char *beg, *end;
    
    beg = Buf;
    for (end = Buf; *end; ++end)
	if (*end == '\n') {
	    pg_line (beg, end);
	    beg = end+1;
	}
}

/* pg_line --- page handleing lives here */

static void
pg_line (char *beg, char *end)
{
    /* is this a new page? */
    if (P.lineno == 0)
	pg_do_header ();
    
    fprintf (P.fp, "%.*s", end-beg+1, beg);
    ++P.lineno;
    
    if (P.lineno >= P.pglen - P.ntail)
	pg_do_tailer ();
}

/* pg_do_tailer --- handle page tail */

static void
pg_do_tailer ()
{
    while (P.lineno < P.pglen) {
	putc ('\n', P.fp);
	++P.lineno;
    }
    
    P.lineno = 0;
}

/* pg_do_header --- header processing lives here */

static void
pg_do_header ()
{
    int i, just, offset;
    char left[256], center[256], right[256], tmp[512];
    char out[512];
    char *tab[3], *p, *t;
    long now;
    
    now = time ((long *) 0);
    tab[0] = left;
    tab[1] = center;
    tab[2] = right;
    for (i = 0; i < P.nhead; i++) {
	if (P.head[i] == NULL) {
	    putc ('\n', P.fp);
	    ++P.lineno;
	    continue;
	}
	memset (out, ' ', P.pgwidth);
	just = 0;
	t = tab[just];
	for (p = P.head[i]; *p; p++) {
	    if (*p != P.sep)
		*t++ = *p;
	    else {
		*t = 0;
		if (just < 2)
		    t = tab[++just];
	    }
	}
	if (t)
	    *t = 0;
	    
	switch (just) {
	case 0:             /* one element, center it */
	    if (*left)
		cftime (tmp, left, &now);
	    else
		strcpy (tmp, left);
	    offset = (P.pgwidth - strlen (tmp)) / 2;
	    copy (out+offset, tmp);
	    break;
	case 1:             /* two elements, left & right justify */
	    if (*left)
		cftime (tmp, left, &now);
	    else
		strcpy (tmp, left);
	    copy (out, tmp);
	    
	    if (*center)
		cftime (tmp, center, &now);
	    else
		strcpy (tmp, center);
	    offset = P.pgwidth - strlen (tmp) - 1;
	    copy (out+offset, tmp);
	    break;
	case 2:             /* three elements, left, center & right */
	    if (*left)
		cftime (tmp, left, &now);
	    else
		strcpy (tmp, left);
	    copy (out, tmp);
	    
	    if (*center)
		cftime (tmp, center, &now);
	    else
		strcpy (tmp, center);
	    offset = (P.pgwidth - strlen (tmp)) / 2;
	    copy (out+offset, tmp);
	    
	    if (*right)
		cftime (tmp, right, &now);
	    else
		strcpy (tmp, right);
	    offset = P.pgwidth - strlen (tmp) - 1;
	    copy (out+offset, tmp);
	    break;
	}
	fprintf (P.fp, "%.*s\n", P.pgwidth, out);
	++P.lineno;
    }
}

/* copy --- copy t to s, WITHOUT the terminator */

static void
copy (register char *s, register char *t)
{
    while (*t)
	*s++ = *t++;
}
